# Copyright 2015-2022 - RoboDK Inc. - https://robodk.com/
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ----------------------------------------------------
# This project is the compiled version of the RoboDK Add-In for Autodesk Fusion 360
# Follow these steps to load the plugin:
#   1- Make sure to install/update RoboDK:
#      https://robodk.com/download
#   2- Unzip the contents of this folder in:
#      C:/RoboDK/Other/Plugin-Fusion360/RoboDK/  
#      (Mac) -> /home/username/RoboDK/Other/Plugin-Fusion360/RoboDK/   
#   3- Open Autodesk Fusion 360
#   4- Select Add-Ins->Scripts and Add-Ins   (in the Model section)
#   5- Select Add-Ins tab
#   6- Select the green "+" button right beside My Add-Ins
#   7- Provide the foder:
#      C:/RoboDK/Other/Plugin-Fusion360/RoboDK/ 
#   8- Select Run
#
# More information about the RoboDK Add-In for Autodesk Fusion 360:
#   https://robodk.com/doc/en/Plugin-Fusion360.html 
#
# More information about the RoboDK API:
#   Documentation: https://robodk.com/doc/en/RoboDK-API.html
#   Reference:     https://robodk.com/doc/en/PythonAPI/index.html
#   GitHub:        https://github.com/RoboDK/RoboDK-API/
#
#------------------------------------------------------

import sys
import os



# Detect Python version and post processor
print("Using Python version: " + str(sys.version_info))
path_app = os.path.dirname(__file__).replace(os.sep,"/")
print("RoboDK Add in for Fusion: " + path_app)

# Check if the post is compatible with the Python version
#version_str = str(sys.version_info[0]) + str(sys.version_info[1])

# Important: Needed on 2022-02-17: We should force loading verion 39.
# For some reason, Autodesk Fusion 360's Python thinks it is still v37 but it is 39
version_str = "39" 
path_library = path_app + '/v' + version_str
if not os.path.isdir(path_library):
    raise Exception("RoboDK Add-In not found. Make sure you are using a supported Python version: " + path_library)

# Load the post processor
exec("from .v" + version_str + ".robodk_autodesk import *")


def run_custom(context):
    ui = getUI()
    try:
        # Auto setup command
        _register_command(Cmd_AutoSetup())
    
        # Command to load the 3D model
        _register_command(Cmd_LoadPart())

        # Command to load curves
        _register_command(Cmd_LoadCurves())

        # Command to load points
        _register_command(Cmd_LoadPoints())
        
        # General settings:
        _register_command(RoboDk())
        
        # Command to load CAM file (NC)
        _register_command(Cmd_LoadCAM())
        
        # Command to post all programs
        _register_command(Cmd_GenerateProg())
 
    except:
        if ui:
            ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))

if __name__== "__main__":
    run()
    